import colorama


class ColorSettings:
    """Settings for color printing

    Attributes:
        is_dev (bool): If developing on IDE, this must be True, otherwise False.
        print_color (bool): For printing colors making use of mcutils print_color must be set to True
    """

    is_dev = False
    print_color = True


class Color:
    """All available colors defined by mcutils"""
    BLACK = '\033[30m'
    RED = '\033[31m'
    GREEN = '\033[32m'
    ORANGE = '\033[33m'
    BLUE = '\033[34m'
    PURPLE = '\033[35m'
    CYAN = '\033[36m'
    LIGHTGREY = '\033[37m'
    DARKGREY = '\033[90m'
    LIGHTRED = '\033[91m'
    LIGHTGREEN = '\033[92m'
    YELLOW = '\033[93m'
    LIGHTBLUE = '\033[94m'
    PINK = '\033[95m'
    LIGHTCYAN = '\033[96m'
    RESET = '\033[0m'


def mcprint(text='', format_='', color=None, end='\n'):
    """print to stdout making use of color formatting options

    Args:
        text (str): Text to be displayed on stdout
        format_ (str): Will add this string at the beginning of the text
        color (Color): Color used for displaying the text
        end (str): Overwrites end arguments of the print() statement
    """
    if not ColorSettings.is_dev:
        colorama.init(convert=True)

    text = '{}{}'.format(format_, text)
    if color and ColorSettings.print_color:
        text = "{}{}{}".format(color, text, Color.RESET)
    print(text, end=end)


def enable_color():
    ColorSettings.print_color = True
    colorama.init()


def disable_color():
    ColorSettings.print_color = False
    colorama.deinit()



from .print_manager import mcprint, Color
import logging


def exit_application(text: str = None, enter_quit: bool =False):
    """Exit application

    Args:
        text (str): Text to display before exiting
        enter_quit (bool): If true, user must press enter before exiting
    """
    if text:
        mcprint(text=text, color=Color.YELLOW)
    logging.info('Exiting Application Code:0')
    if enter_quit:
        get_input('Press Enter to exit...')
    exit(0)


def print_error(operators_list=None, contains_list=None, return_type=None):
    if operators_list:
        for operator in operators_list:
            if return_type == int:
                logging.warning('input must be {}'.format(operator))
            elif return_type == str:
                logging.warning('input length must be {}'.format(operator))
    if contains_list:
        logging.warning('input must be one of the following')
        for contains in contains_list:
            mcprint(text='\t{}'.format(contains), color=Color.RED)


def input_validation(user_input, return_type, valid_options):
    if return_type == int:
        if not user_input.isnumeric():
            return False
        user_input = int(user_input)

    # Contains validation
    if valid_options:

        operators_list = list(filter(lambda x: str(x).startswith(('<', '>', '==', '!=')), valid_options))
        contains_list = list(set(valid_options) - set(operators_list))

        # Complex validation
        # Special operators
        for operator in operators_list:
            if '<=' in operator:
                value = operator.replace('<=', '')
                if return_type == int:
                    if not user_input <= int(value):
                        print_error(operators_list=operators_list, return_type=return_type)
                        return False
                elif return_type == str:
                    if not len(user_input) <= int(value):
                        print_error(operators_list=operators_list, return_type=return_type)
                        return False

            elif '>=' in operator:
                value = operator.replace('>=', '')
                if return_type == int:
                    if not user_input >= int(value):
                        print_error(operators_list=operators_list, return_type=return_type)
                        return False
                elif return_type == str:
                    if not len(user_input) >= int(value):
                        print_error(operators_list=operators_list, return_type=return_type)
                        return False

            elif '<' in operator:
                value = operator.replace('<', '')
                if return_type == int:
                    if not user_input < int(value):
                        print_error(operators_list=operators_list, return_type=return_type)
                        return False
                elif return_type == str:
                    if not len(user_input) < int(value):
                        print_error(operators_list=operators_list, return_type=return_type)
                        return False

            elif '>' in operator:
                value = operator.replace('>', '')
                if return_type == int:
                    if not user_input > int(value):
                        print_error(operators_list=operators_list, return_type=return_type)
                        return False
                elif return_type == str:
                    if not len(user_input) > int(value):
                        print_error(operators_list=operators_list, return_type=return_type)
                        return False

            elif '==' in operator:
                value = operator.replace('==', '')
                if return_type == int:
                    if not user_input == int(value):
                        print_error(operators_list=operators_list, return_type=return_type)
                        return False
                elif return_type == str:
                    if not len(user_input) == int(value):
                        print_error(operators_list=operators_list, return_type=return_type)
                        return False
            elif '!=' in operator:
                value = operator.replace('!=', '')
                if return_type == int:
                    if not user_input != int(value):
                        print_error(operators_list=operators_list, return_type=return_type)
                        return False
                elif return_type == str:
                    if not len(user_input) != int(value):
                        print_error(operators_list=operators_list, return_type=return_type)
                        return False

        # if contains in valid options
        if len(contains_list) > 0:
            if user_input not in contains_list:
                return False

    return True


def get_input(format_: str = '>> ', text: str = None, can_exit: bool = True, exit_input: str= 'exit',
              valid_options: list = None, return_type: type = str,
              validation_function=None, color: Color = None):
    """
    Require the user to input a value

    Args:
        format_ (str): Will add this string at the beginning of the text
        text (str): Text to be displayed before input
        can_exit (bool): If true, when user types the exit_input command, the application will exit
        exit_input (str): Special string, when user inputs this command, the application will exit
        valid_options (list): If the input of the user is not in valid_options, the menu will ask for it again
        return_type (type): Indicates the required return type
        validation_function (function): Function used to validate if the input is correct
        color (Color): Color used for displaying the text
    """
    if text:
        mcprint(text=text, color=color)

    while True:
        user_input = input(format_)

        # Emergency exit system
        if user_input == exit_input:
            if can_exit:
                exit_application()
            else:
                logging.warning('Can\'t exit application now')

        # This is the build-in validations system
        if validation_function:
            validation = validation_function.__call__(user_input)

        # This is the external validation system
        else:
            # from input_validation import input_validation
            validation = input_validation(user_input=user_input, return_type=return_type, valid_options=valid_options)
        if validation:
            break

        logging.warning('Not Valid Entry')

    return user_input



import os
import logging
from datetime import datetime
from .print_manager import mcprint, Color


class DirectoryManager:
    class File:
        def __init__(self, path, name, extension, size, created_at):
            self.path = path
            self.name = name
            self.extension = extension
            self.size = size
            self.created_at = created_at

        def print_info(self):
            mcprint("Name: {}".format(self.name))
            mcprint("Path: {}".format(self.path))
            mcprint("Extension: {}".format(self.extension))
            mcprint("Size: {}".format(self.size))
            mcprint("Created at: {}".format(self.created_at))
            print()

        # Modify delete function
        def delete_file(self):
            mcprint("Deleting File <{}>".format(self.path), color=Color.RED)
            os.remove(self.path)

    def __init__(self, directories=None):
        if not directories:
            directories = [os.getcwd()]
        self.directories = directories
        self.files = []
        self.selected_files = []
        self.get_files()

    def get_dirs(self):
        dirs_list = []
        for file in self.files:
            dirs_list.append(file.path)
        return dirs_list

    # Retrieves a list of Files in self.files
    def get_files(self):
        import os

        def create_file(directory_name, new_file_name=None):

            file_dir = directory_name
            if new_file_name:
                file_dir = os.path.join(directory_name, new_file_name)
            else:
                new_file_name = file_dir.rsplit('\\', 1)[-1]

            created_at = datetime.fromtimestamp(os.path.getctime(file_dir)).strftime('%Y-%m-%d %H:%M:%S')
            file = self.File(file_dir, new_file_name, new_file_name.rsplit('.', 1)[-1],
                             os.path.getsize(file_dir), created_at)
            self.files.append(file)

        for directory in self.directories:
            if os.path.isdir(directory):
                if os.path.exists(directory):
                    for file_name in os.listdir(directory):
                        create_file(directory, file_name)
                else:
                    logging.error("Path \"{}\" doesn't exists".format(directory))
            elif os.path.isfile(directory):
                create_file(directory_name=directory)
            else:
                logging.error("Path \"{}\" not found".format(directory))

    def print_files_info(self):
        for file in self.files:
            file.print_info()

    def filter_format(self, extensions=None):
        new_files = []
        for file in self.files:
            if file.extension in extensions:
                new_files.append(file)
        self.files = new_files

    @staticmethod
    def create_directory(directory):
        try:
            os.makedirs(directory)
        except IsADirectoryError:
            logging.error(error_string="Couldn't create the directory '{}'".format(directory))

    def open_file(self, file):
        import platform
        import subprocess
        current_os = platform.system()

        if isinstance(file, self.File):
            path = file.path
        elif isinstance(file, str):
            path = file
        else:
            raise NotADirectoryError

        if os.path.isfile(path):
            logging.info("Open File <{}> // current os {}".format(file, current_os))
            if current_os == 'Linux':
                subprocess.call(('xdg-open', path))
            elif current_os == 'Windows':
                os.startfile(path)
            elif current_os == "Darwin":
                subprocess.call(('open', path))
            else:
                logging.error("OS not supported")

        else:
            logging.error("File \"{}\" not found".format(path))

    def add_file_to_selection(self, *args):
        logging.info("Adding Files <{}> to Selection".format(args))
        files = None
        for arg in args:
            if isinstance(arg, self.File):
                files = [arg]
            elif isinstance(arg, list):
                files = list(arg)
            elif isinstance(arg, str):
                files = list(filter(lambda x: arg in x.name, self.files))
            if files:
                self.selected_files += files
        return self.selected_files

    def clear_file_selection(self):
        self.selected_files.clear()



import json


def get_dict_from_json(path):
    """
    Loads json file as dict

    Args:
        path (str): path of the json file to load
    """
    with open(path, "r") as json_file:
        dictionary = json.load(json_file)
    return dictionary


def generate_json(path, dictionary):
    """
    Writes dict as json

    Args:
        path (str): path to write the json file
        dictionary (dict): dict to write as json
    """
    with open(path, "w") as json_file:
        json.dump(dictionary, json_file)



from .directory_manager import *
from .input_validation import *
from .print_manager import *
from .json_manager import *
from .utilities import *
from .menu import *
from .logger import *

"""Top-level package for mcutils."""

__author__ = """Matias Canepa Gonzalez"""
__email__ = 'macanepa@miuandes.cl'
__version__ = '1.1.1'

"""Main module."""


disable_mc_logger()
enable_color()



from .print_manager import mcprint
from .input_validation import get_input
from .utilities import clear
import logging


class MenuFunction:
    """Allows to assign a user defined function to the Menu

    If the function requires arguments, these can be defined while creating
    a MenuFunction. This MenuFunction must be placed inside the options argument in Menu

    Args:
        title (str): Name to display on the Menu
        function (function): Function to be called once it is selected on the Menu

    Example:
        The following example showcases how to define a new ``MenuFunction``
        using a user defined function which requires arguments

        >>> import mcutils as mc
        ...
        >>> def foo(n):
        ...     return n**2
        ...
        >>> mf_foo = mc.MenuFunction(title='do foo',
        ...                          function=foo,
        ...                          n=4)

        The example above, defines a ``MenuFunction`` which has attached
        a user defined function ``foo(n)``

    """
    def __init__(self, title=None, function=None, **kwargs):
        self.function = function
        self.title = title
        self.kwargs = kwargs
        self.returned_value = None

    def print_function_info(self):
        mcprint('Function: %s' % self.function)

        for parameter in self.kwargs:
            mcprint(parameter)

    def get_unassigned_params(self):
        unassigned_parameters_list = []
        for parameter in self.function.func_code.co_varnames:
            if parameter not in self.kwargs:
                mcprint(parameter)
                unassigned_parameters_list.append(parameter)
        return unassigned_parameters_list

    def get_args(self):
        mcprint(self.kwargs)
        return self.kwargs

    def call_function(self):
        self.returned_value = self.function(**self.kwargs)
        return self.returned_value


class Menu:
    """Main structure for generating menus

    Args:
        title (str): Title of the menu
        subtitle (str): Subtitle of the menu
        text (str): Text of the menu
        options (list): list or dict of the available options to select from
        return_type (type): Indicates the required return type
        parent (Menu): The parent of the current menu. Used to head back to the previous menu
        input_each (bool): If true, the menu will ask the user to input the value for each of the options. Returns a dict
        back (bool): If true and adittional option with index 0 will appear in the menu. Use this to head back to the previous Menu.

    Attributes:
        return_value (object): If input_each = False, returns the selected index of the Menu. If input_each = True
                            returns a dictionary containing the options as keys and the values introduced by the user
        function_returned_value (object): Returns the returned by the user defined function from MenuFunction

    Example:
        Create simple Menu

        >>> import mcutils as mc
        >>> mc_menu = mc.Menu(title='Main Menu',
        ...                   subtitle='Subtitle',
        ...                   text='Please select one of the following options',
        ...                   options=['Option 1', 'Option 2', 'Option 3'])
        >>> mc_menu.show()
        ...
        === Main Menu
        - - Subtitle
        ...
        Please select one of the following options
        1. Option 1
        2. Option 2
        3. Option 3
        0. Back
        >>


    Example:
        Create Menu using user defined functions and submenus

        The following example explains how to display other menus as
        submenus, and call user defined functions
        >>> import mcutils as mc
        >>> def foo(n):
        ...     return n**2
        ...
        >>> mf_foo = mc.MenuFunction(title='do foo', function=foo, n=4)
        ...
        >>> mc_submenu = mc.Menu(title='Submenu',
        ...                      text='This is the submenu',
        ...                      options=[mf_foo])
        ...
        >>> mc_menu = mc.Menu(title='Main Menu',
        ...                   subtitle='Subtitle',
        ...                   text='Please select one of the following options',
        ...                   options=[mc_submenu, 'Option 2', 'Option 3'])
        ...
        >>> mc_menu.show()
        >>> print(mc_submenu.function_returned_value)

        As shown above, two menus were defined. One for the main menu ``mc_menu``,
        and the one for the submenu ``mc_submenu``. A user defined function was defined aswell
        as ``mf_foo``. The submenu was added into the options of main menu, while the user defined
        function was assigned to the submenu.

    """
    def __init__(self, title: str = None, subtitle: str = None, text: str = None,
                 options=None, return_type: type = int, parent=None,
                 input_each: bool = False,
                 previous_menu=None, back: bool = True):
        self.title = title
        self.subtitle = subtitle
        self.text = text
        self.options = options
        self.return_type = return_type
        self.parent = parent
        self.input_each = input_each
        self.previous_menu = previous_menu
        self.back = back
        self.returned_value = None
        self.function_returned_value = None

    def set_parent(self, parent):
        self.parent = parent

    def set_previous_menu(self, previous_menu):
        self.previous_menu = previous_menu

    def get_selection(self):

        start_index = 1
        if self.back:
            start_index = 0

        # if there exist options it means user have to select one of them
        if (self.options.__len__() != 0) and (not self.input_each):

            while True:

                selection = get_input()

                if selection.__str__().isdigit():
                    if int(selection) in range(start_index, (self.options.__len__()) + 1):
                        if int(selection) != 0:
                            if isinstance(self.options[int(selection) - 1], MenuFunction):
                                function = self.options[int(selection) - 1]
                                self.function_returned_value = function.call_function()
                            elif isinstance(self.options[int(selection) - 1], Menu):
                                sub_menu = self.options[int(selection) - 1]
                                sub_menu.set_parent(self)
                                sub_menu.show()
                        else:
                            if self.parent:
                                self.parent.set_previous_menu(self)
                                self.parent.show()
                        break
                    else:
                        logging.warning('Index not in range')

                else:
                    logging.warning('Entered value must be int')

        elif self.input_each:
            selection = {}
            for option in self.options:
                if isinstance(self.options, dict):
                    filter_criteria = self.options[option]
                    return_type = int
                    if filter_criteria[0] in [str, int]:
                        return_type = filter_criteria[0]
                        filter_criteria = filter_criteria[1:]
                    parameter_value = get_input(format_='{} >> '.format(option),
                                                valid_options=filter_criteria,
                                                return_type=return_type)
                else:
                    parameter_value = get_input('{} >> '.format(option))
                selection[option] = parameter_value

        # if there aren't any option it means user must input a string
        else:
            selection = get_input()

        self.returned_value = selection
        return selection

    def show(self):
        """Display the menu once

        The Menu will be displayed until the required operations are completed"""
        # if(self.previous_menu != None) and (self != self.previous_menu):
        #     del(self.previous_menu)
        clear()
        if self.title:
            mcprint('=== %s ' % self.title)
        if self.subtitle:
            mcprint('- - %s' % self.subtitle)
        print()
        if self.text:
            mcprint(self.text)

        # print 'Parent:',self.parent
        if self.options and not self.input_each:
            for index, option in enumerate(self.options):
                if isinstance(option, MenuFunction):
                    print('%s. %s' % (str(index + 1), option.title))
                elif isinstance(option, Menu):
                    print('%s. %s' % (str(index + 1), option.title))
                else:
                    print('%s. %s' % (str(index + 1), option))
            if self.back:
                mcprint('0. Back')

        selected_option = self.get_selection()
        return selected_option




from .print_manager import mcprint, Color
from .input_validation import get_input
from datetime import datetime
import logging


class About:
    def __init__(self,
                 authors=None,
                 company_name=None,
                 team_name=None,
                 github_account=None,
                 email_address=None,
                 github_repo=None):

        self.authors = authors
        self.company_name = company_name
        self.team_name = team_name
        self.github_account = github_account
        self.github_repo = github_repo
        self.email_address = email_address

    def print_credits(self):
        clear(100)
        mcprint('>> About <<')
        if self.company_name:
            mcprint('Company: {}'.format(self.company_name))
        if self.team_name:
            mcprint('Developed by {}'.format(self.team_name))
        if self.authors:
            mcprint('\nAuthors:')
            for author in self.authors:
                mcprint('\t-{}'.format(author))
        print()
        if self.email_address:
            mcprint('Email: {}'.format(self.email_address))
        if self.github_account:
            mcprint('GitHub: {}'.format(self.github_account))
        if self.github_repo:
            mcprint('GitHub Repository: {}'.format(self.github_repo))
        input('\nPress Enter to Continue...')


def clear(n=3):
    print('\n' * n)


def date_generator(include_time=False, year=None, month=None, day=None, hour=None, minute=None, second=None):
    """Generate a datetime object defined by the user

    This function will generate a datetime object. For each attribute not defined,
    the user must fill it in. If the user leaves a blank attribute, the current date & time
    will be used instead. The datetime can include the time if include_time is set to True

    Args:
        include_time (bool): If True, the user will be required to input hour, minute and second
        year (int): year
        month (int): month, must be in range [1, 12]
        day (int): day, must be in range of the corresponding month
        hour (int): hour, must be in range [0, 23]
        second (int): second, must be in range [0, 59]
    """
    if not year:
        while True:
            try:
                year = get_input(format_='Year: ')
                if year == '':
                    year = datetime.now().year
                year = int(year)
                datetime(year, 1, 1)
                break
            except ValueError:
                logging.warning('Enter a valid year')

    if not month:
        while True:
            try:
                month = get_input(format_='Month: ')
                if month == '':
                    month = datetime.now().month
                month = int(month)
                datetime(year, month, 1)
                break
            except ValueError:
                logging.warning('Enter a valid month')

    if not day:
        while True:
            try:
                day = get_input(format_='Day: ')
                if day == '':
                    day = datetime.now().day
                day = int(day)
                datetime(year, month, day)
                break
            except ValueError:
                logging.warning('Enter a valid day')

    if not include_time:
        date = datetime(year, month, day)
        return date

    if not hour:
        while True:
            try:
                hour = get_input(format_='Hour: ')
                if hour == '':
                    hour = datetime.now().hour
                hour = int(hour)
                datetime(year, month, day, hour)
                break
            except ValueError:
                logging.warning('Enter a valid hour')

    if not minute:
        while True:
            try:
                minute = get_input(format_='Minute: ')
                if minute == '':
                    minute = datetime.now().minute
                minute = int(minute)
                datetime(year, month, day, hour, minute)
                break
            except ValueError:
                logging.warning('Enter a valid minute')

    if not second:
        while True:
            try:
                second = get_input(format_='Second: ')
                if second == '':
                    second = datetime.now().second
                second = int(second)
                datetime(year, month, day, hour, minute, second)
                break
            except ValueError:
                logging.warning('Enter a valid second')
    date = datetime(year, month, day, hour, minute, second)
    return date
